/*
Odbiornik IR
Ty-grysek, 2020.01

Nacinij krtko przycisk aby wczy/wyczy zasilanie bez pilota.

Procedura parowania z dowolnym pilotem IR.
Nacinij i przytrzymaj przez 2 sek. przycisk - dioda LED zacznie byska pojedynczym byskiem, jeli zasilanie byo wczone - wyczy si.
W cigu 60 sek. skieruj pilot w kierunku odbiornika IR i nacinij na pilocie przycisk ktry ma wcza zasilanie.
Dioda LED zacznie byska podwjnym byskiem, zasilanie wczy si. W cigu 60 sek. skieruj pilot w kierunku odbiornika IR i nacinij na pilocie przycisk ktry ma wycza zasilanie.
Przycisk pilota wczajcy i wyczajcy zasilanie moe by jednym i tym samym przyciskiem.
Zakoczenie nauki potwierdzone jest dugim (ok. 1s) zawieceniem si diody LED.

Kasowanie nauczonych kodw pilota IR: przytrzymaj przycisk przez 8 sek.
*/

/*
Fusebity:  
  BOOTSZ	= 1024W_0C00
  BOOTRST	= [ ]
  RSTDISBL	= [ ]
  DWEN		= [ ]
  SPIEN		= [X]
  WDTON		= [ ]
  EESAVE	= [X]
  BODLEVEL	= DISABLED
  CKDIV8	= [ ]
  CKOUT		= [ ]
  SUT_CKSEL	= INTRCOSC_8MHZ_6CK_14CK_4MS1

  EXTENDED	= 0xF9 (valid)
  HIGH		= 0xD7 (valid)
  LOW		= 0xD2 (valid)
*/


#define F_CPU 8000000UL		//8MHz wewntrzny oscylator

#include <avr/io.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/eeprom.h>
#include "irmp.c"

#define RELPORT			PORTD
#define RELPORT_dir		DDRD
#define RELPORT_pin		PIND
#define REL_pin			_BV(PORTD4)

#define LEDPORT			PORTC
#define LEDPORT_dir		DDRC
#define LEDPORT_pin		PINC
#define LED_pin			_BV(PORTC2)

#define BTNPORT			PORTD
#define BTNPORT_dir		DDRD
#define BTNPORT_pin		PIND
#define BTN_pin			_BV(PORTD3)

#define IRPORT			PORTD
#define IRPORT_dir		DDRD
#define IRPORT_pin		PIND
#define IR_pin			_BV(PORTD2)

#define LernTime		60		//po tym czasie [s] nauka programowania zostanie przerwana
#define BtnTimeLern		2000	//po tym czasie [ms] dugiego nacinicia przycisku zostanie uruchomiony proces nauki kodw IR
#define BtnTimeMax		8000	//po tym czasie [ms] dugiego nacinicia przycisku zostan skasowane nauczone kody IR
#define SleepTime		30000	//po tym czasie bezczynnoci [ms] nastpuje upienie mikrokontrolera

uint8_t RelState;		//stan przekanika: 0 / 1
uint8_t mode;			//0: gotowo, 1: nauka IR wczenia, 2: nauka IR wyczenia
uint16_t ms;			//Licznik ms
uint16_t LernDelay;		//odliczanie czasu [s] do zakoczenia trybu nauki
IRMP_DATA irmp_ON;		//kod pilota IR do wczenia
IRMP_DATA irmp_OFF;		//kod pilota IR do wyczenia
IRMP_DATA i0, i1;		//tymczasowe na czas nauki
uint16_t BtnTime;		//liczy czas dugiego nacinicia przycisku


// ###################################################################################################
// ###                                            IRMP                                             ###
// ###################################################################################################

//https://www.mikrocontroller.net/articles/IRMP_-_english

#ifndef F_CPU
	#error F_CPU unknown
#endif

//timer1
static void timerIRMP_init (void)
{
	#if defined (__AVR_ATtiny45__) || defined (__AVR_ATtiny85__)				// ATtiny45 / ATtiny85:
		#if F_CPU >= 16000000L
			OCR1C   =  (F_CPU / F_INTERRUPTS / 8) - 1;                          // compare value: 1/15000 of CPU frequency, presc = 8
			TCCR1   = (1 << CTC1) | (1 << CS12);                                // switch CTC Mode on, set prescaler to 8
		#else
			OCR1C   =  (F_CPU / F_INTERRUPTS / 4) - 1;                          // compare value: 1/15000 of CPU frequency, presc = 4
			TCCR1   = (1 << CTC1) | (1 << CS11) | (1 << CS10);                  // switch CTC Mode on, set prescaler to 4
		#endif
	#else                                                                       // ATmegaXX:
		OCR1A   =  (F_CPU / F_INTERRUPTS) - 1;                                  // compare value: 1/15000 of CPU frequency
		TCCR1B  = (1 << WGM12) | (1 << CS10);                                   // switch CTC Mode on, set prescaler to 1
	#endif
	
	#ifdef TIMSK1
		TIMSK1  = 1 << OCIE1A;                                                  // OCIE1A: Interrupt by timer compare
	#else
		TIMSK   = 1 << OCIE1A;                                                  // OCIE1A: Interrupt by timer compare
	#endif
}

//timer1
void timerIRMP_stop()
{
	#if defined (__AVR_ATtiny45__) || defined (__AVR_ATtiny85__)
		TCCR1 = 0;
	#else
		TCCR1B = 0;  //wycz rdo licznika
	#endif
}

#ifdef TIM1_COMPA_vect                                                      // ATtiny84
	#define COMPA_VECT  TIM1_COMPA_vect
#else
	#define COMPA_VECT  TIMER1_COMPA_vect                                   // ATmega
#endif

ISR(COMPA_VECT)                                                             // Timer1 output compare A interrupt service routine, called every 1/15000 sec
{
	(void) irmp_ISR();                                                      // call irmp ISR
	// call other timer interrupt routines...
}

//porwnuje 2 kody IR, zwraca 1 jeli rwne lub 0 jeli rne
uint8_t ir_compare(IRMP_DATA ir0, IRMP_DATA ir1)
{
	return (ir0.protocol == ir1.protocol) &&
	(ir0.address == ir1.address) &&
	(ir0.command == ir1.command);
}

// ###################################################################################################
// ###                                   STEROWANIE PERYFERIAMI                                    ###
// ###################################################################################################


//zapala lub gasi diod LED
void LED(uint8_t stan)
{
	if (stan) LEDPORT |= LED_pin;
	else LEDPORT &= ~LED_pin;
}

//wcza lub wycza przekanik
void REL(uint8_t stan)
{
	if (stan) RELPORT |= REL_pin;
	else RELPORT &= ~REL_pin;
	if (stan) RelState = 1;
	else RelState = 0;
}

//sprawdza stan przycisku
uint8_t Button()
{
	if (!(BTNPORT_pin & BTN_pin)) return 1;
	else return 0;
}

// ###################################################################################################
// ###                           TIMER2 - licznik i obsuga diody LED                              ###
// ###################################################################################################

//wywoanie co 1ms
ISR(TIMER2_COMPA_vect)
{
	ms++;
	
	if (mode)
	{
		if (ms > 1000)
		{
			ms = 0;
			LernDelay++;
			if (LernDelay > LernTime)
			{
				mode = 0;
				LED(0);
				REL(0);
				irmp_ON = i0;
				irmp_OFF = i1;
				LernDelay = 0;
			}
		}
		else
		{
			if (ms == 1) LED(1);
			else if (ms == 50) LED(0);
			if (mode == 2)
			{
				if (ms == 150) LED(1);
				else if (ms == 200) LED(0);			
			}
		}
	}
	
	if (BtnTime) BtnTime++;
}

void Timer2Start()
{
	ms = 0;
	LernDelay = 0;
	OCR2A  = 125;						//grna warto licznika
	TIMSK2 |= _BV(OCIE2A);				//przerwanie przy porwnaniu z OCR2A
	TCCR2A |= _BV(WGM21);				//tryb 2 CTC
	TCCR2B |= _BV(CS22);				//tryb 2 CTC, rdo licznika = CLK/64, start
}

void Timer2Stop()
{
	TCCR2B = 0;  //wycz rdo licznika
}

// ###################################################################################################
// ###                                          SLEEP                                              ###
// ###################################################################################################

//wywoanie przerwa ma tylko wybudzi mikrokontroler
ISR(INT0_vect)
{
	
}

ISR(INT1_vect)
{
	
}

//inicjuje parametry po wybudzeniu
void WakeUp()
{
	EIMSK &= ~(_BV(INT0) | _BV(INT1));			//zabro przerwa INT0 i INT1
	timerIRMP_init();
	Timer2Start();
}

void GoSleep()
{
	timerIRMP_stop();
	EIMSK |= _BV(INT0) | _BV(INT1);				//zezwl na przerwania
	sleep_enable();
	//sleep_bod_disable();
	sleep_cpu();

	sleep_disable();
	WakeUp();
}

// ###################################################################################################
// ###                                           EEPROM                                            ###
// ###################################################################################################

//zapisuje do EEPROM kody pilota IR
void EEPROM_write()
{
	//protocol - 1 byte
	//address - 2 bytes
	//command - 2 bytes
	uint16_t addr = 0;
	eeprom_write_block((void*)&irmp_ON,(void*)addr,5);
	addr = 5;
	eeprom_write_block((void*)&irmp_OFF,(void*)addr,5);
}

//odczytuje z EEPROM kody pilota IR
void EEPROM_read()
{
	uint16_t addr = 0;
	eeprom_read_block((void*)&irmp_ON,(void*)addr,5);
	addr = 5;
	eeprom_read_block((void*)&irmp_OFF,(void*)addr,5);
}

// ###################################################################################################
// ###                                         MAIN                                                ###
// ###################################################################################################

int main(void)
{
	IRMP_DATA   irmp_data;			//biece dane z pilota IR

	//inicjalizacja portw
	RELPORT &= ~REL_pin;			//stan niski
	RELPORT_dir |= REL_pin;			//ustaw jako wyjcie
	LEDPORT |= LED_pin;				//stan wysoki - dioda zapalona
	LEDPORT_dir |= LED_pin;			//ustaw jako wyjcie
	IRPORT_dir &= ~IR_pin;			//ustaw jako wejcie
	IRPORT &= ~IR_pin;				//wycz pull-up rezystor
	BTNPORT_dir &= ~BTN_pin;		//ustaw jako wejcie
	BTNPORT |= BTN_pin;				//wcz pull-up rezystor

	//bynij LED na dzie dobry	
	_delay_ms(200);
	EEPROM_read();
	LED(0);

	//inicjalizacja
	MCUCR &= ~(_BV(ISC01) | _BV(ISC11));							//przerwanie dla INT0 i INT1 przy zboczu opadajcym
	set_sleep_mode(SLEEP_MODE_PWR_DOWN);							//ustawienie trybu upienia
	sei();
	irmp_init();
	WakeUp();
	mode = 0;
	BtnTime = 0;
	
    while (1) 
    {
		
		if (irmp_get_data (&irmp_data))						//odczytano kod pilota IR
		{
			if (!(irmp_data.flags & IRMP_FLAG_REPETITION))	//tylko pierwsze nacinicie przycisku pilota
			{
				if (mode)				//tryb nauki
				{
					Timer2Stop();				
					if (mode == 1)					
					{
						REL(1);
						irmp_ON = irmp_data;
						mode++;
					}
					else		//zakocz tryb nauki
					{
						REL(0);
						LED(1);
						irmp_OFF = irmp_data;
						EEPROM_write();
						_delay_ms(800);
						LED(0);
						mode = 0;						
					}
					Timer2Start();
				}
				else	//tryb oczekiwania
				{
					if ((RelState == 0) && (ir_compare(irmp_data,irmp_ON)))
					{
						ms = 0;
						REL(1);
						LED(1);
						_delay_ms(50);
						LED(0);
					}
					else if ((RelState == 1) && (ir_compare(irmp_data,irmp_OFF)))
					{
						ms = 0;
						REL(0);
						LED(1);
						_delay_ms(50);
						LED(0);
					}
				}
			}
		}
		
		
		if (Button())
		{
			if (!BtnTime)
			{
				//krtkie nacinicie - wcz/wycz
				BtnTime = 1;
				ms = 0;
				LED(1);
				if (RelState == 0) REL(1);
				else REL(0);
				_delay_ms(50);	//drgania stykw + bysk diody LED
				LED(0);
			}
			else if ((BtnTime > BtnTimeLern) && (mode == 0))
			{
				//rozpocznij nauk kodw IR
				REL(0);
				//zapamitaj aktualne kody IR na wypadek przerwania procesu nauki
				i0 = irmp_ON;
				i1 = irmp_OFF;
				mode = 1;		//tryb nauki
				ms = 0;
			}
			else if (BtnTime > BtnTimeMax)
			{
				//kasowanie nauczonych kodw IR
				mode = 0;
				ms = 0;
				LED(1);
				irmp_ON.protocol = 0xFF;
				irmp_ON.address = 0xFFFF;
				irmp_ON.command = 0xFFFF;
				irmp_OFF.protocol = 0xFF;
				irmp_OFF.address = 0xFFFF;
				irmp_OFF.command = 0xFFFF;
				EEPROM_write();
				while (Button());	//czekaj na zwolnienie przycisku
				LED(0);
			}
		}
		else if (BtnTime)
		{
			_delay_ms(20);	//drgania stykw
			BtnTime = 0;
		}
		
		if ((!mode) && (ms > SleepTime)) GoSleep();		//po 30 sek. bezczynnoci w trybie czuwania upij
    }
}